#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifndef UNICODE_CONVERT
#include <iconv.h>
#endif



/* Counting the number of blocks classified by the
   maxiumum UTF-8 sequence length within the block. */
#ifdef BLOCK_COUNTING
int block_counts[4] = {0,0,0,0};
#endif

#define BUFFER_SIZE 8096

#ifdef UNICODE_ORG_CONVERT
#undef STD_ICONV
#undef FAST_U8U16
#define error_in_result(rslt) rslt != conversionOK
#define illegal_sequence_error(rslt) rslt == sourceIllegal
#define incomplete_sequence_error(rslt) rslt == sourceExhausted
#endif

#ifndef UNICODE_ORG_CONVERT
#define error_in_result(rslt) rslt == (size_t) -1
#define illegal_sequence_error(rslt) errno == EILSEQ
#define incomplete_sequence_error(rslt) errno == EINVAL


#ifdef STD_ICONV
#undef FAST_U8U16
#endif
#ifndef STD_ICONV
#define FAST_U8U16 1
#endif
#endif

// Include the conversion routines.
#ifdef UNICODE_ORG_CONVERT
//#include "Unicode.org/ConvertUTF.h"
#include "Unicode.org/ConvertUTF.c"
#endif
#ifdef FAST_U8U16
#include "libu8u16.c"
#endif


void do_UTF16toUTF8(FILE *infile, FILE *outfile) {


  unsigned short UTF8_buffer[BUFFER_SIZE] __attribute__((aligned(16)));
  unsigned char UTF16_buffer[BUFFER_SIZE] __attribute__((aligned(16)));
  unsigned short * srcbuf_ptr = &UTF8_buffer[0];
  unsigned char * trgtbuf_ptr = &UTF16_buffer[0];
  size_t inbytes_left, outbytes_left;
  int file_position = 0;

  unsigned int chars_read, i, UTF16_chars;
    


   iconv_t cd = iconv_open("UTF-8", "UTF-16BE");
   size_t rslt;


  chars_read = fread(&UTF8_buffer, 1, BUFFER_SIZE, infile);

  while (chars_read > 0) {
    srcbuf_ptr = &UTF8_buffer[0];
    trgtbuf_ptr = &UTF16_buffer[0];
    inbytes_left = chars_read;
/*    outbytes_left = BUFFER_SIZE*2;   */
    outbytes_left = BUFFER_SIZE ;   
#ifdef STD_ICONV
    rslt = iconv(cd,
                 (char **) &srcbuf_ptr, 
                 &inbytes_left,
                 (char **) &trgtbuf_ptr, 
                 &outbytes_left);
#endif
#ifdef UNICODE_ORG_CONVERT
    rslt = ConvertUTF8toUTF16((const UTF8**) &srcbuf_ptr, 
                              srcbuf_ptr + chars_read,
                              &trgtbuf_ptr, 
                              &UTF16_buffer[BUFFER_SIZE],
                              strictConversion);
    inbytes_left = chars_read - (srcbuf_ptr - &UTF8_buffer[0]);
#endif


    file_position += chars_read - inbytes_left;
/*    UTF16_chars = ((int) trgtbuf_ptr - (int) &UTF16_buffer[0])/2;   */
    UTF16_chars = ((int) trgtbuf_ptr - (int) &UTF16_buffer[0]);   
/*     fwrite(&UTF16_buffer, 2, UTF16_chars, outfile);   */
     fwrite(&UTF16_buffer, 1, UTF16_chars, outfile);
    if (error_in_result(rslt)) {
      if (illegal_sequence_error(rslt)) {
        int pos = 0;
        fprintf(stderr, "Illegal UTF-8 sequence at position %i in source.\n", file_position);
        close(infile);
        close(outfile);
        exit(-1);
      }
      if (!incomplete_sequence_error(rslt)) {
        fprintf(stderr, "Unknown error %i at position %i in source.\n", errno, file_position);
        close(infile);
        close(outfile);
        exit(-1);
      }
      // errno == EINVAL  or rslt == sourceExhausted
      // Incomplete sequence at end of input buffer. 
      if (chars_read < BUFFER_SIZE) {
        int pos = 0;
        fprintf(stderr, "EOF with incomplete UTF-8 sequence at position %i in source.\n", file_position);

        close(infile);
        close(outfile);
        exit(-1);
      }
      // Move unprocessed characters to beginning.
      for (i = 0; i < inbytes_left; i++) {
        UTF8_buffer[i] = UTF8_buffer[chars_read - inbytes_left + i];
      }

    }
    chars_read = fread(&UTF8_buffer[inbytes_left], 1, BUFFER_SIZE-inbytes_left, infile);
    chars_read += inbytes_left;
  }
#ifdef STD_ICONV
  iconv_close(cd);
#endif
  close(infile);
  close(outfile);
}




int
main(int argc, char * argv[]) {
  if (argc < 2) {
    printf("Usage: %s <filename.u16> [<filename.u8>].\n", argv[0]);
          exit(-1);
  }
  char * filename = argv[1];
  FILE *infile, *outfile;
  infile = fopen(filename, "rb");
  if (!infile) {
      fprintf(stderr, "Error: cannot open %s for input.\n", filename);
      exit(-1);
  }

  if (argc < 3) outfile = stdout;
  else {
    outfile = fopen(argv[2], "wb");
    if (!outfile) {
      fprintf(stderr, "Error: cannot open %s for writing.\n", argv[2]);
      exit(-1);
    }
  }

  do_UTF16toUTF8(infile, outfile);

  return(0);
}
